{-
	Funktionale Programmierung WS2010
	Implementierung des Boyer-Moore Algorithmus in Haskell (nach Richard Bird)

	Der Boyer-Moore Algorithmus ist ein String-Matching-Algorithmus und wurde
	1977 von Robert S. Boyer und J Strother Moore entwickelt.
	Er wird meistens mithilfe zweier heuristischer Regeln realisiert,
	der Bad Character Regel und der Good Suffix Regel.
	
	Beide Regeln werden in diesem Lsungsansatz nicht explizit verwendet.
	Die hier vorgeschlagene Realisierung des BM-Algorithmus leitet sich 
	vielmehr aus den symbolischen Umformungen und effizienzsteigernden 
	Gesetzen (scan lemma, tupling law of foldl) ab, die auf einen anfangs 
	schlechten String-Matching-Algorithmus angwendet werden. 	
	Die Schlsselidee des BM-Algorithmus, nmlich das pattern-matching
	von rechts nach links, taucht hier in der entsprechenden Definition
	der Funktion endswith auf. Weitere Optimierungen sollen schlielich
	zu einer BM-Realisierung mit linearer Laufzeit fhren.
	
	Autoren: 
		Tobias Lders
		Pascal Rhode
-}

module Main where

import Data.List
import System.IO
import System.CPUTime
import Text.Printf
import Data.Array
import System.Environment

--------------------------------------------
-- Beispiele zum Aufrufen (fr Prsentation)	
--------------------------------------------
		
matches0 = do
			t1 <- getCPUTime
			x <- readFile "supersuppe.txt"
			let res1 = processMatches0 "supersuppe" x
			putStrLn ("Matchings : " ++ show res1) 
			t2 <- getCPUTime
			let diff = (fromIntegral (t2-t1)) / (10^12)
			printf "Computation time: %0.3f sec\n" (diff :: Double)
			
matches1 = do
			t1 <- getCPUTime
			x <- readFile "supersuppe.txt"
			let res1 = processMatches1 "supersuppe" x
			putStrLn ("Matchings : " ++ show res1) 
			t2 <- getCPUTime
			let diff = (fromIntegral (t2-t1)) / (10^12)
			printf "Computation time: %0.3f sec\n" (diff :: Double)
			
matches2 = do
			t1 <- getCPUTime
			x <- readFile "supersuppe.txt"
			let res1 = processMatches2 "supersuppe" x
			putStrLn ("Matchings : " ++ show res1) 
			t2 <- getCPUTime
			let diff = (fromIntegral (t2-t1)) / (10^12)
			printf "Computation time: %0.3f sec\n" (diff :: Double)
			
matches3 = do
			t1 <- getCPUTime
			x <- readFile "supersuppe.txt"
			let res1 = processMatches3 "supersuppe" x
			putStrLn ("Matchings : " ++ show res1)  
			t2 <- getCPUTime
			let diff = (fromIntegral (t2-t1)) / (10^12)
			printf "Computation time: %0.3f sec\n" (diff :: Double)
	
----------------------------
-- Hilfsfunktionen	
----------------------------

--prft, ob ein String mit einem Pattern endet (rckwrts)
endswith	:: Eq a => [a] -> [a] -> Bool
endswith ws = (reverse ws #).reverse	

-- Prfix-Relation
[] # vs 		= True
(u:us) # [] 	= False
(u:us) # (v:vs) = (u == v && us # vs)	

-- Bestimmt die Lnge des lngsten gemeinsamen Prfixes 
-- zweier Strings
llcp				:: Eq a => [a] -> [a] -> Int
llcp xs []			= 0
llcp [] ys			= 0
llcp (x:xs) (y:ys) 	= if x==y then 1+llcp xs ys else 0	
	
-- Schrittweite fr Suchalgorithmus 
-- (Schrittweite von 1 <=> ein Buchstabe weiterspringen)
step		 ::(Int,[a]) -> a -> (Int,[a])
step(n,sx) x = (n+1,x:sx)	

-- Alle gemeinsamen Prfixe
allcp		:: Eq a => [a] -> [Int]
allcp xs 	= [llcp xs(drop k xs) | k <- [0..length xs-1]]	
allcp' xs 	= tail(allcp xs) ++ [0]
	
-------------------------------
-- String-Matching Algorithmen
-------------------------------

-- naiver Ansatz mit teurem inits
processMatches0    :: Eq a => [a] -> [a] -> [Int]
processMatches0 ws = map length.filter(endswith ws).inits

-- Basis-Implementierung des Boyer-Moore Algorithmus
processMatches1    :: Eq a => [a] -> [a] -> [Int]
processMatches1 ws = map fst . filter((sw #) . snd) . scanl step(0,[])
					where 
					sw = reverse ws

-- Erste Optimierung des Boyer-Moore, 
-- allerdings mit teurer Shift-Berechnung
processMatches2    :: Eq a => [a] -> [a] -> [Int]
processMatches2 ws = test m . scanl step(0,[])
					where
					test j [] = []
					test j ((n,sx) : nxs) 
						| i == m 	= n : test k (drop(k-1)nxs)
						| m-k <= i	= test k (drop(k-1)nxs)
						| otherwise	= test m (drop(k-1)nxs)
							where
							i'	= llcp sw (take j sx)	
							i	= if i' == j then m else i'
							k	= shift sw i
								where
								shift sw i = head[k|k <- [1..m], llcp sw (drop k sw) == min i (m-k)]
					(sw,m)	= (reverse ws, length ws)				

-- Zweite Optimierung des Boyer-Moore 
-- mit indiziertem (konstanten) Array-Zugriff
processMatches3    :: Eq a => [a] -> [a] -> [Int]
processMatches3 ws = test m . scanl step(0,[])
					where
					test j [] = []
					test j ((n,sx) : nxs) 
						| i == m 	= n : test k (drop(k-1)nxs)
						| m-k <= i	= test k (drop(k-1)nxs)
						| otherwise	= test m (drop(k-1)nxs)
							where
							i'	= llcp sw (take j sx)	
							i	= if i' == j then m else i'
							k	= accumArray min m (0,m) (vks ++ vks')!i
								where
								m 			= length sw														
								vks 		= zip(allcp' sw) [1..m]					
								vks'		= zip[m,m-1..1](foldr op [] vks)
								op(v,k) ks 	= if v+k == m then k:ks else head ks:ks
					(sw,m)	= (reverse ws, length ws)

	









